"""Module for keep all filtering functionalities

"""
import logging

import pandas as pd
import numpy as np
import xarray as xr

module_logger = logging.getLogger("lidarwind.filters")


def filter_status(ds: xr.Dataset):
    """Filter dataset based on WindCube's software

    Parameters
    ----------
    ds: xr.Dataset
        Dataset with LIDAR records

    Returns
    -------

    xr.Dataset:
        A Dataset with records where status was true (1)
    """
    if "radial_wind_speed_status90" not in ds:
        module_logger.error(
            "filter_status() requires radial_wind_speed_status90"
        )
        raise ValueError

    return ds.where(ds.radial_wind_speed_status90 == 1)


def filter_snr(ds: xr.Dataset, snr: float):
    if "cnr90" not in ds:
        module_logger.error("filter_snr() requires cnr90")
        raise ValueError

    return ds.where(ds.cnr90 > snr)


class Filtering:
    """SNR and Status filter

    It uses the carrier to noise ratio (SNR) and status
    variables available in the WindCube's data. It is
    similar to the filter described in the manual

    Parameters
    ----------
    data : xrarray.Dataset
        Dataset containing the original WindCube's data

    """

    def __init__(self, data):

        self.data = data

    def get_vertical_obs_comp(self, variable, snr=False, status=True):
        """Vertical data filter

        It uses the SNR and status variables to filter out
        the data from vertical observations

        Parameters
        ----------
        variable : str
            name of the variable that will be filtered

        snr : bool, int, optional
            if an interger is given it is used to
            as threshold to filter the data based on
            the signal to noise ratio

        status : bool, optional
            if true it filters the data using the status
            variable generated by the WindCube's software

        Returns
        -------
        tmp_data : xarray.DataSet
            an instance of the variable filtered using
            SNR or status variable

        """

        tmp_data = self.data[variable]

        if status:
            tmp_data = tmp_data.where(
                self.data.radial_wind_speed_status90 == 1
            )

        if snr is not False:
            tmp_data = tmp_data.where(self.data.cnr90 > snr)

        tmp_data = tmp_data.where(self.data.elevation == 90, drop=True)

        return tmp_data

    def get_radial_obs_comp(self, variable, azm, snr=False, status=True):
        """Slanted data filter

        It uses the SNR and status variables to filter out
        the data from slanted observations

        Parameters
        ----------
        variable : str
            name of the variable that will be filtered

        snr : bool, int, optional
            if an interger is given it is used to
            as threshold to filter the data based on
            the signal to noise ratio

        status : bool, optional
            if true it filters the data using the status
            variable generated by the WindCube's software

        Returns
        -------
        tmpaData : xarray.DataSet

            an instance of the variable filtered using
            SNR or status variable

        """

        tmp_data = self.data[variable]

        if status:
            tmp_data = tmp_data.where(self.data.radial_wind_speed_status == 1)

        if snr is not False:
            tmp_data = tmp_data.where(self.data.cnr > snr)

        tmp_data = tmp_data.where(
            (self.data.elevation != 90) & (self.data.azimuth == azm), drop=True
        )

        return tmp_data


# it removes the STE below cloud layer
class SecondTripEchoFilter:
    """Boundary layer second trip echoes filter

    This filter minimises the presence of second
    trip echoes (STE). This filter is based on the
    standard deviation of the anomaly of the observaions.
    It is applicable in regions where there is a contrast
    between the real data and the STE.

    Parameters
    ----------
    data : object
        the object returned from the GetRestructuredData

    timeCloudMaks : xarray.DataArray
        it is a time series for indicating the presence
        of clouds above the maximum WinCube range.
        1 indicates cloud and 0 indicates no cloud.
        (THIS MAKS IS NOT NEEDED NOW)

    n_prof : int
        number of profiles used to calculating the anomaly

    center : bool, optional
        it defines how the mean value for the anomaly will
        be calculated

    min_periods : int
        minimum number of profiles used for calculating the
        mean value

    n_std : int
        Multiplication factor for defining the size of the
        window to keep the data. The filter removes any
        anomaly larger than n_std * std

    str_h : str
        starting hour for calculating the anomaly

    end_h : str
        end hour for calculating the anomaly

    Returns
    -------
    object : object
        an object containing data filtered for STE

    """

    def __init__(
        self,
        data,
        n_prof=500,
        center=True,
        min_periods=30,
        n_std=2,
        str_h="09",
        end_h="16",
    ):

        self.lidar = data
        # self.time_cloud_mask = time_cloud_mask
        self.n_prof = n_prof
        self.center = center
        self.min_periods = min_periods
        self.n_std = n_std

        self.get_time_edges(str_h=str_h, end_h=end_h)
        self.cal_mean_and_anom_slant()
        self.cal_mean_and_anom_90()
        self.cleaning()
        self.cleaning90()

    def get_time_edges(self, str_h="09", end_h="16"):
        """
        It creates the time boundaries for the STD anomaly calculation

        Parameters
        ----------
        str_h : str
            starting hour for calculating the anomaly

        end_h : str
            end hour for calculating the anomaly

        """
        sel_time = pd.to_datetime(self.lidar.data_transf.time.values[0])
        sel_time = sel_time.strftime("%Y%m%d")
        self.start_time = pd.to_datetime(f"{sel_time} {str_h}")
        self.end_time = pd.to_datetime(f"{sel_time} {end_h}")

    def cal_mean_and_anom_slant(self):
        """
        It calculates the anomaly from the slanted observations
        """

        # slanted beam
        tmp_sel_data = self.lidar.data_transf

        self.data_mean = tmp_sel_data.rolling(
            time=self.n_prof, center=self.center, min_periods=self.min_periods
        ).mean()

        self.data_anom = self.lidar.data_transf - self.data_mean

    def cal_mean_and_anom_90(self):
        """
        It calculates the anomaly from the vertical observations
        """

        # vertical beam
        tmp_sel_data_90 = self.lidar.data_transf_90

        self.data_mean90 = tmp_sel_data_90.rolling(
            time=self.n_prof, center=self.center, min_periods=self.min_periods
        ).mean()

        self.data_anom_90 = self.lidar.data_transf_90 - self.data_mean90

    def cleaning(self):
        """
        It removes the data that is larger than the n_std * anomaly
        from the slanted observations
        """

        tmp_anom = self.data_anom.where(
            (self.lidar.data_transf.time > self.start_time)
            & (self.lidar.data_transf.time < self.end_time)
        )

        anom_std = tmp_anom.std(dim=["time", "range", "elv"])

        tmp_clean_data = self.lidar.data_transf.copy()
        tmp_clean_data = tmp_clean_data.where(
            np.abs(self.data_anom) < self.n_std * anom_std
        )
        self.lidar.data_transf.values = tmp_clean_data.values

    def cleaning90(self):
        """
        It removes the data that is larger than the n_std * anomaly
        from the vertical observations
        """

        tmp_anom = self.data_anom_90.where(
            (self.lidar.data_transf_90.time > self.start_time)
            & (self.lidar.data_transf_90.time < self.end_time)
        )

        anom_std = tmp_anom.std(dim=["time", "range90"])

        tmp_clean_data = self.lidar.data_transf_90.copy()
        tmp_clean_data = tmp_clean_data.where(
            np.abs(self.data_anom_90) < self.n_std * anom_std
        )

        self.lidar.data_transf_90.values = tmp_clean_data.values


# it removes STE and clouds contamination
# from above the aerosol loaded region
class WindCubeCloudRemoval:
    """Above boundary layer second trip echoes filter

    This filter reduces the second trip echoes contamination
    and clouds from regions above the boundary layer. It
    requires information from the ceilometer.
    IT IS STILL EXPERIMENTAL

    Parameters
    ----------
    ceilo : xarray.Dataset
        A dataset of the CHM15k Nimbus ceilometer observations.

    lidar : xarray.Dataset
        An instance of the re-structured WindCube dataset

    Returns
    -------
    object : object
        an object containing an instance of the noise
        height interface and the re-structured dataset
        filtered for STE and clouds.

    """

    def __init__(self, ceilo, lidar=None):

        self.lidar = lidar
        self.ceilo = ceilo

        self.get_noise_free_beta()
        self.get_height_interface()

        if lidar is not None:
            self.get_interp_interf_height()
            self.remove_cloud()

    def get_noise_free_beta(self):
        """
        It removes the noise from the backscattered signal
        """
        positive_beta = self.ceilo.beta_raw.where(self.ceilo.beta_raw > 0)

        positive_beta = positive_beta.rolling(
            range=10, center=True, min_periods=10
        ).mean()
        positive_beta = positive_beta.rolling(
            time=15, center=True, min_periods=15
        ).mean()

        self.noise_free_beta = positive_beta

        return self

    def get_height_interface(self):
        """
        It identifies the height of the separation between
        the noise and the non-noise data from the ceilometer
        backscattered signal
        """
        positive_beta = self.ceilo.beta_raw.where(self.ceilo.beta_raw > 0)

        tmp_ceilo_hgt = positive_beta.copy()
        tmp_values = tmp_ceilo_hgt.values
        tmp_values[np.isfinite(tmp_values)] = 1
        tmp_ceilo_hgt.values = tmp_values
        tmp_ceilo_hgt = tmp_ceilo_hgt * self.ceilo.range

        lowest_beta = self.noise_free_beta.where(tmp_ceilo_hgt < 4e3)

        tmp_ceilo_hgt = tmp_ceilo_hgt.where(np.isfinite(lowest_beta))

        self.interf_height = tmp_ceilo_hgt.max(dim="range")
        self.interf_height = self.interf_height.rolling(
            time=7, center=True, min_periods=5
        ).mean()

        return self

    def get_interp_interf_height(self):
        """
        It interpolates the noise height interface to the same
        temporal resolution from the windcube data
        """
        self.interp_interf_height = self.interf_height.interp(
            time=self.lidar.data_transf.time
        )
        self.interp_interf_height_90 = self.interf_height.interp(
            time=self.lidar.data_transf_90.time
        )

        return self

    def remove_cloud(self):
        """
        It removes from the windcube's observation all
        data above the noise height interface
        """

        tmp_height = self.lidar.data_transf.copy()
        tmp_values = tmp_height.values
        tmp_values[np.isfinite(tmp_values)] = 1
        tmp_height.values = tmp_values
        tmp_height = tmp_height * self.lidar.data_transf.range
        self.lidar.data_transf = self.lidar.data_transf.where(
            tmp_height < self.interp_interf_height
        )

        tmp_height = self.lidar.data_transf_90.copy()
        tmp_values = tmp_height.values
        tmp_values[np.isfinite(tmp_values)] = 1
        tmp_height.values = tmp_values
        tmp_height = tmp_height * self.lidar.data_transf_90.range90
        self.lidar.data_transf_90 = self.lidar.data_transf_90.where(
            tmp_height < self.interp_interf_height_90
        )
        self.lidar.relative_beta90 = self.lidar.relative_beta90.where(
            tmp_height < self.interp_interf_height_90
        )

        return self



import pytest
import xarray as xr
import numpy as np

import lidarwind as lst


def get_dummy_dbs():

    wind = (
        np.sin(np.deg2rad(np.array([0, 180, 270, 90])))
        * 1
        / np.sin(np.deg2rad(45))
    )

    elv = np.array([45, 45, 45, 45, 90])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 180, 270, 90, 0])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.append(wind, 0)[:, np.newaxis],
        dims=("time", "gate_index"),
        coords={"time": np.arange(len(elv)), "gate_index": [1]},
    )

    data_status = xr.DataArray(
        np.array([1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "gate_index"),
        coords={"time": np.arange(len(elv)), "gate_index": [1]},
    )

    data_mean_time = xr.DataArray(
        np.array([1, 1, 1, 1, 1]),
        dims=("time"),
        coords={"time": np.arange(len(elv))},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr": data_status,
            "measurement_height": data_status,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "scan_mean_time": data_mean_time,
        }
    )

    return test_ds


@pytest.fixture
def get_wind_profiles():
    return lst.GetWindProperties5Beam(get_dummy_dbs())


def test_get_wind_properties_5_beam_vert_wind_dim(get_wind_profiles):
    assert "gate_index" not in get_wind_profiles.ver_wind_speed.dims


def test_get_wind_properties_5_beam_comp_u_dim(get_wind_profiles):
    assert "gate_index" not in get_wind_profiles.comp_u.dims


def test_get_wind_properties_5_beam_comp_v_dim(get_wind_profiles):
    assert "gate_index" not in get_wind_profiles.comp_v.dims


def test_get_wind_properties_5_beam_comp_u_value(get_wind_profiles):
    assert np.round(get_wind_profiles.comp_u.values, 1) == -2.0


def test_get_wind_properties_5_beam_comp_v_value(get_wind_profiles):
    assert np.round(get_wind_profiles.comp_v.values, 1) == 0


def test_get_wind_properties_5_beam_wind_speed_value(get_wind_profiles):
    assert np.round(get_wind_profiles.hor_wind_speed.values, 1) == 2


def test_get_wind_properties_5_beam_wind_dir_value(get_wind_profiles):
    assert np.round(get_wind_profiles.hor_wind_dir.values, 1) == 270


def test_get_wind_properties_5_beam_vert_wind_value(get_wind_profiles):
    assert np.round(get_wind_profiles.ver_wind_speed.values, 1) == 0



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from .filters import Filtering


class PlotSettings:
    def __init__(self, mpl, style="dark_background"):

        self.mpl = mpl
        self.style = style

    def update_settings(self):

        font_size = 16

        self.mpl.style.use(self.style)
        self.mpl.rcParams["figure.figsize"] = [6, 6]
        self.mpl.rcParams["figure.dpi"] = 80
        self.mpl.rcParams["savefig.dpi"] = 100

        self.mpl.rcParams["font.size"] = font_size
        self.mpl.rcParams["legend.fontsize"] = font_size
        self.mpl.rcParams["figure.titlesize"] = font_size

        self.mpl.rcParams["ytick.labelsize"] = font_size
        self.mpl.rcParams["xtick.labelsize"] = font_size
        self.mpl.rcParams["axes.titlesize"] = font_size
        self.mpl.rcParams["axes.labelsize"] = font_size

        self.mpl.rcParams["legend.fancybox"] = True
        self.mpl.rcParams["legend.framealpha"] = 0.7
        self.mpl.rcParams["legend.facecolor"] = "silver"
        self.mpl.rcParams["legend.frameon"] = True

        self.mpl.rcParams["lines.linewidth"] = 5

        return self

    def plot_setup(plot):

        plt.setp(plot.axes.xaxis.get_majorticklabels(), rotation=0)
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        plot.axes.xaxis.set_major_formatter(formatter)

        plt.grid(b=True)

        return plot


class Visualizer:
    def __init__(self, data):

        self.data = data

    def view_orig_var(
        self,
        var_name,
        cmap="Spectral",
        vmin=-1,
        vmax=1,
        elv="90",
        azm="-",
        save=False,
        plot_id=None,
        fig_path=None,
        name_prefix=None,
        show=False,
        min_time=None,
        max_time=None,
    ):

        if plot_id == "rad_wind_speed_panel":

            tmp_data = self.data
            self.plot_data_azm(
                data_non_90=tmp_data,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                plot_id=plot_id,
                fig_path=fig_path,
                save=save,
                show=show,
                min_time=min_time,
                max_time=max_time,
            )

        else:
            tmp_data = Filtering(self.data).get_vertical_obs_comp(var_name)

            if name_prefix:
                std_name = tmp_data.attrs["standard_name"]
                string_name = f"{name_prefix}_{std_name}"

            else:
                string_name = tmp_data.attrs["standard_name"]

            self.plot_data(
                tmp_data=tmp_data,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                elv=elv,
                azm=azm,
                save=save,
                plot_id=plot_id,
                string_name=string_name,
                fig_path=fig_path,
                min_time=min_time,
                max_time=max_time,
            )

    def view_ret_var(
        self,
        var_name,
        cmap="Spectral",
        vmin=-1,
        vmax=1,
        elv="90",
        azm="-",
        save=False,
        plot_id=None,
        fig_path=None,
        name_prefix=None,
        show=False,
        min_time=None,
        max_time=None,
    ):

        tmp_data = self.data[var_name]

        string_name = tmp_data.attrs["standard_name"]

        self.plot_data(
            tmp_data=tmp_data,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            elv=elv,
            azm=azm,
            save=save,
            plot_id=plot_id,
            string_name=string_name,
            fig_path=fig_path,
            show=show,
            min_time=min_time,
            max_time=max_time,
        )

    def plot_data(
        self,
        tmp_data,
        cmap="Spectral",
        vmin=-1,
        vmax=1,
        elv="90",
        azm="-",
        save=False,
        plot_id=None,
        fig_path=None,
        string_name=None,
        show=False,
        min_time=None,
        max_time=None,
    ):

        sel_day = pd.to_datetime(tmp_data.time[0].values)

        if max_time is not None:
            max_time = pd.to_datetime(max_time)

        else:
            max_time = pd.to_datetime(sel_day.strftime("%Y%m%d 23:59:59"))

        if min_time is not None:
            min_time = pd.to_datetime(min_time)

        else:
            min_time = pd.to_datetime(sel_day.strftime("%Y%m%d 00:00:00"))

        tmp_data = tmp_data.sel(time=slice(min_time, max_time))

        if string_name:
            tmp_data.attrs["standard_name"] = string_name

        plt.figure(figsize=(18, 8))
        plot = tmp_data.plot(x="time", cmap=cmap, vmin=vmin, vmax=vmax)
        plot = PlotSettings.plot_setup(plot)

        plt.grid(b=True)
        plt.ylim(0, 12e3)
        plt.xlim(min_time, max_time)
        plt.title(f"elv: {elv}, azm: {azm}")

        if plot_id == "hor_wind_dir":
            plot.colorbar.set_ticks(np.linspace(0, 360, 9))

        if save:
            sel_day_str = sel_day.strftime("%Y%m%d")
            file_name = f"{sel_day_str}_{plot_id}.png"
            output_file_name = os.path.join(fig_path, file_name)
            print(output_file_name)
            plt.savefig(output_file_name, bbox_inches="tight")

        if show:
            plt.show()

        plt.close()

    def plot_data_azm(
        self,
        data_non_90,
        cmap="Spectral",
        vmin=-1,
        vmax=1,
        fig_path=None,
        save=False,
        plot_id=None,
        show=False,
        min_time=None,
        max_time=None,
    ):

        elv = data_non_90.elv.values[0]
        fig, axes = plt.subplots(5, 1, sharex=True, figsize=(18, 25))

        sel_day = pd.to_datetime(data_non_90.time[0].values)

        if max_time is not None:
            max_time = pd.to_datetime(max_time)

        else:
            max_time = pd.to_datetime(sel_day.strftime("%Y%m%d 23:59:59"))

        if min_time is not None:
            min_time = pd.to_datetime(min_time)

        else:
            min_time = pd.to_datetime(sel_day.strftime("%Y%m%d 00:00:00"))

        for ax_number, i in enumerate(data_non_90.azm.values):

            tmp_data = data_non_90.sel(azm=i)

            tmp_data = tmp_data.sel(time=slice(min_time, max_time))

            plot = tmp_data.plot(
                x="time", cmap=cmap, vmin=vmin, vmax=vmax, ax=axes[ax_number]
            )

            plot = PlotSettings.plot_setup(plot)

            axes[ax_number].grid(b=True)
            axes[ax_number].set_ylim(0, 12e3)
            axes[ax_number].set_xlim(min_time, max_time)
            axes[ax_number].set_title(f"elv: {elv}, azm: {i}")

        if save:
            sel_day_str = sel_day.strftime("%Y%m%d")
            file_name = f"{sel_day_str}_{plot_id}.png"
            output_file_name = os.path.join(fig_path, file_name)
            print(output_file_name)
            plt.savefig(output_file_name, bbox_inches="tight")

        if show:
            plt.show()

        plt.close()



import logging

import numpy as np
import xarray as xr
import xrft

from .data_attributes import LoadAttributes
from .data_operator import GetRestructuredData

module_logger = logging.getLogger("lidarwind.wind_prop_retrieval")
module_logger.debug("loading wind_prop_retrieval")


class FourierTransfWindMethod:
    """FFT wind retrieval method

    It is a fft wind retrieval method. It was proposed
    by Ishwardat (2017). For more details see
    http://resolver.tudelft.nl/uuid:a659654b-e76a-4513-a656-ecad761bdbc8

    Parameters
    ----------
    doppler_obs : xarray.DataArray
        It should be a DataArray of slanted Doppler velocity
        observations as function of the azimuthal angle.
        It must have a coordinate called azm.

    Returns
    -------
    wind_prop : xarray.Dataset
        A Dataset containing the horizontal wind speed
        and direction. It also includes the zonal and
        meridional wind components

    """

    def __init__(self, doppler_obs: xr.DataArray):

        self.logger = logging.getLogger(
            "lidarwind.wind_prop_retrieval.FourierTransfWindMethod"
        )
        self.logger.info("creating an instance of FourierTransfWindMethod")

        if not isinstance(doppler_obs, xr.DataArray):
            self.logger.error("wrong data type: expecting a xr.DataArray")
            raise TypeError

        self.doppler_obs = doppler_obs

        self.get_comp_amp()
        self.get_phase()
        self.get_rad_wind_speed()
        self.get_hor_wind_speed()
        self.get_wind_dir()
        self.get_wind_comp_u()
        self.get_wind_comp_v()

    def get_comp_amp(self):
        """First harmonic amplitude

        It calculates the complex amplitudes from the
        first harmonic from the observations along the
        azm coordinate
        """

        self.logger.info("calculating the complex amplitude")

        self.comp_amp = xrft.fft(
            self.doppler_obs, dim=["azm"], true_amplitude=False
        ).isel(freq_azm=-2)

        return self

    def get_phase(self):
        """First harmonic phase

        It calculates the phase of the first harmonic from the complex
        amplitude used for retrieving wind direction, i.e. comp_amp.
        """

        self.logger.info("calculating the phase from the complex amplitude")

        self.phase = -np.rad2deg(
            np.arctan2(self.comp_amp.imag, self.comp_amp.real)
        )

        self.phase.attrs = {
            "long_name": "Retrived phase",
            "units": "degree",
            "comments": "Phase derived from comp_amp "
            "variable using the FFT method.",
        }

        return self

    def get_wind_dir(self):
        """Wind direction calculation

        It calculates the wind direction based on the
        phase of the first harmonic
        """

        self.logger.info("retrieving wind direction from the phase")

        self.wind_dir = self.phase + 180
        self.wind_dir.attrs = {
            "long_name": "Wind direction",
            "standard_name": "wind_from_direction",
            "units": "degree",
            "comments": "Wind direction retrived using the FFT method.",
        }

        return self

    def get_rad_wind_speed(self):
        """Wind speed calculation

        It calculates the wind speed using the first harmonic
        """

        self.logger.info(
            "calculating the radial wind speed from the complex amplitude"
        )

        self.rad_wind_speed = (
            2 * np.abs(self.comp_amp) / self.doppler_obs.azm.shape[0]
        )

        self.rad_wind_speed.attrs = {
            "long_name": "wind speed not corrected for the elevation",
            "units": "m s-1",
            "comments": "Radial wind velocity retrived using "
            "the FFT method not corrected for the elevation.",
        }

        return self

    def get_hor_wind_speed(self):
        """Wind speed elevation correction

        It corrects the magnitude of the wind speed using
        the elevation angle
        """

        self.logger.info("retrieving the horizontal wind speed")

        self.hor_wind_speed = self.rad_wind_speed / np.cos(
            np.deg2rad(self.doppler_obs.elv)
        )

        self.hor_wind_speed.attrs = {
            "long_name": "horizontal wind speed",
            "standard_name": "wind_speed",
            "units": "m s-1",
            "comments": "Horizontal wind velocity retrived using "
            "the FFT method corrected for the elevation.",
        }

        return self

    def get_azm_wind(self, azm):
        """Wind speed for a given azimuth

        It retrieves the wind speed for a given azimuthal angle.
        It cam be used to calculate the meridional and zonal wind
        components

        Parameters
        ----------
        azm : float
            an azimuth for retrieving the wind
        """

        self.logger.info("calculating wind speed for a give azimuth")

        azm_hor_wind = self.rad_wind_speed * np.sin(
            np.deg2rad(azm) + np.deg2rad(self.phase.values + 180)
        )
        azm_hor_wind = azm_hor_wind / np.cos(np.deg2rad(self.doppler_obs.elv))

        return azm_hor_wind

    def get_wind_comp_u(self):
        """Zonal wind calculation

        It retrives the zonal wind component
        the -1 multiplication is for comply with the
        conventions used in meteorology
        """

        self.logger.info("retrieving the zonal wind speed component")

        self.comp_u = self.get_azm_wind(0) * -1
        self.comp_u.name = "comp_u"
        self.comp_u.attrs = {
            "long_name": "Zonal wind component",
            "units": "m s-1",
            "comments": "Zonal wind component retrieved using the FFT method.",
        }

        return self

    def get_wind_comp_v(self):
        """Meridional wind calculation

        It retrieves the meridional wind component
        the -1 multiplication is for comply with the
        conventions used in meteorology
        """

        self.logger.info("retrieving the meridional wind speed component")

        self.comp_v = self.get_azm_wind(90) * -1
        self.comp_v.name = "comp_v"
        self.comp_v.attrs = {
            "long_name": "Meridional wind component",
            "units": "m s-1",
            "comments": "Meridional wind component "
            "retrieved using the FFT method.",
        }

        return self

    def wind_prop(self):
        """Wind dataset

        It creates and returnes a dataset containing
        the wind speed, direction, and components
        """

        self.logger.info(
            "creating a xarray dataset from the retrieved wind properties"
        )

        wind_prop = xr.Dataset()
        wind_prop["horizontal_wind_direction"] = self.wind_dir
        wind_prop["horizontal_wind_speed"] = self.hor_wind_speed
        wind_prop["zonal_wind"] = self.comp_u
        wind_prop["meridional_wind"] = self.comp_v

        return wind_prop


class GetWindProperties5Beam:
    """DBS wind retrieval

    This class caculates the wind speeed and direction
    using the 5bean dataset (DBS files) as input.

    Parameters
    ----------
    data : xarray.Dataset
        merged xarray dataset (merged_ds) output from
        lidarwind.DbsOperations()

    status_filter : bolean
        Data filtering based on the wind lidar
        wind status variable. If True, all data with status not
        equal to 1 are removed. If False, no filtering is applied.

    cnr : int, optional
        Filter based on the carrier to noise ratio.
        If None, no filtering is applied. If a cnr value is given,
        all data smaller than the cnr is removed.

    method : str
        It can be 'single_dbs' (default) or 'continuous'.
        If single_dbs: the wind information is retrieved
        from comple sets of DBS profiles. If continuos:
        it uses the nearest 4 observations to retrieve
        the wind information, a tolerance wind has to
        be specified.

    tolerance : str
        It defines the tolerance window that the method will
        use for identify the nearest profile. Example: '8s'
        for 8 seconds.

    Returns
    -------
    object : object
        This class returns an object containing the
        derived wind speed (.hor_wind_speed) and
        direction (.hor_wind_dir).

    """

    def __init__(
        self,
        data: xr.Dataset,
        status_filter=True,
        cnr=None,
        method="single_dbs",
        tolerance="8s",
    ):

        self.logger = logging.getLogger(
            "lidarwind.wind_prop_retrieval.GetWindProperties5Beam"
        )
        self.logger.info("creating an instance of GetWindProperties5Beam")

        if not isinstance(data, xr.Dataset):
            self.logger.error("wrong data type: expecting a xr.Dataset")
            raise TypeError

        if status_filter:
            data["radial_wind_speed"] = data.radial_wind_speed.where(
                data.radial_wind_speed_status == 1
            )

        if cnr is not None:
            data["radial_wind_speed"] = data.radial_wind_speed.where(
                data.cnr >= cnr
            )

        elevation = data.elevation.round(1)

        time90 = elevation.time.where(elevation == 90, drop=True)
        time_non_90 = elevation.time.where(elevation != 90, drop=True)

        azimuth_non_90 = data.azimuth.sel(
            time=time_non_90, method="Nearest"
        ).round(1)
        azimuth_non_90[azimuth_non_90 == 360] = 0

        self.tolerance = tolerance

        self.azimuth_non_90 = azimuth_non_90
        self.elevation_non_90 = elevation.sel(time=time_non_90)

        # replace range by measurement_height
        self.range_val_non_90 = data.measurement_height.sel(time=time_non_90)
        self.rad_wind_speed_non_90 = data.radial_wind_speed.sel(
            time=time_non_90
        )
        self.mean_time_non_90 = data.scan_mean_time.sel(time=time_non_90)

        self.range_val_90 = data.measurement_height.sel(time=time90)
        self.ver_wind_speed = data.radial_wind_speed.sel(time=time90)
        self.correct_vert_wind_comp()

        if method == "continuous":
            self.calc_hor_wind_comp_continuous()

        if method == "single_dbs":
            self.calc_hor_wind_comp_single_dbs()

        self.calc_hor_wind_speed()
        self.calc_hor_wind_dir()

    def correct_wind_comp(self, comp):
        """
        This function replaces the gate_index coordinate
        by the measurement_height.
        (For any component)

        Parameters
        ----------
        comp : xarray.DataArray
            a variable from the original dataset

        Returns
        -------
        comp : xarray.DataArray
            updated variable

        """

        self.logger.info(
            f"replacing the gate_index coordinate by range: {comp.name}"
        )

        comp = comp.rename({"gate_index": "range"})
        comp = comp.assign_coords({"range": self.range_val_90.values[0]})
        comp.range.attrs = self.range_val_90.attrs

        return comp

    def correct_vert_wind_comp(self):
        """
        This function replaces the original coordinate from the vertical
        wind component by the measurement_height.
        """

        self.logger.info("replacing coordinate from the vertical measurement")

        self.ver_wind_speed.name = "compW"
        ver_wind_speed = self.correct_wind_comp(self.ver_wind_speed)
        self.ver_wind_speed = ver_wind_speed

        return self

    def calc_hor_wind_comp_single_dbs(self):
        """
        This method derives v and u components from the
        WindCube DBS files. The components are caculated
        from each individual DBS file. The mean time from each
        scan complete scan is used as identification tag.
        """

        self.logger.info(
            "calculating the horizontal wind using the SINGLE DBS method"
        )

        comp_wind_speed = self.rad_wind_speed_non_90 / (
            2 * np.cos(np.deg2rad(self.elevation_non_90))
        )

        comp_vn = comp_wind_speed.where(self.azimuth_non_90 == 0, drop=True)
        mean_time_vn = self.mean_time_non_90.where(
            self.azimuth_non_90 == 0, drop=True
        )
        comp_vn = comp_vn.assign_coords({"time": mean_time_vn})

        comp_vs = comp_wind_speed.where(self.azimuth_non_90 == 180, drop=True)
        mean_time_vs = self.mean_time_non_90.where(
            self.azimuth_non_90 == 180, drop=True
        )
        comp_vs = comp_vs.assign_coords({"time": mean_time_vs})

        comp_ue = comp_wind_speed.where(self.azimuth_non_90 == 90, drop=True)
        mean_time_ue = self.mean_time_non_90.where(
            self.azimuth_non_90 == 90, drop=True
        )
        comp_ue = comp_ue.assign_coords({"time": mean_time_ue})

        comp_uw = comp_wind_speed.where(self.azimuth_non_90 == 270, drop=True)
        mean_time_uw = self.mean_time_non_90.where(
            self.azimuth_non_90 == 270, drop=True
        )
        comp_uw = comp_uw.assign_coords({"time": mean_time_uw})

        self.comp_v = -(comp_vn - comp_vs)
        self.comp_u = -(comp_ue - comp_uw)

        self.comp_v.name = "comp_v"
        self.comp_u.name = "comp_u"

        self.comp_v = self.correct_wind_comp(self.comp_v)
        self.comp_u = self.correct_wind_comp(self.comp_u)

    def calc_hor_wind_comp_continuous(self):
        """
        Function to derive wind v and u components.
        It folows the same approach used by the lidar software.
        """

        self.logger.info(
            "calculating the horizontal wind using the CONTINUOUS DBS method"
        )

        comp_wind_speed = self.rad_wind_speed_non_90 / (
            2 * np.cos(np.deg2rad(self.elevation_non_90))
        )

        self.comp_vn = comp_wind_speed.where(
            self.azimuth_non_90 == 0, drop=True
        )
        self.comp_vs = comp_wind_speed.where(
            self.azimuth_non_90 == 180, drop=True
        )
        comp_vs = self.comp_vs.reindex(
            time=self.comp_vn.time, method="Nearest", tolerance=self.tolerance
        )

        self.comp_ue = comp_wind_speed.where(
            self.azimuth_non_90 == 90, drop=True
        )
        self.comp_uw = comp_wind_speed.where(
            self.azimuth_non_90 == 270, drop=True
        )
        comp_uw = self.comp_uw.reindex(
            time=self.comp_ue.time, method="Nearest", tolerance=self.tolerance
        )

        self.comp_v = -(self.comp_vn - comp_vs)
        self.comp_u = -(self.comp_ue - comp_uw)

        self.comp_v.name = "comp_v"
        self.comp_u.name = "comp_u"

        self.comp_v = self.correct_wind_comp(self.comp_v)
        self.comp_u = self.correct_wind_comp(self.comp_u)

        self.comp_u = self.comp_u.reindex(
            time=self.comp_v.time, method="Nearest", tolerance=self.tolerance
        )

        return self

    def calc_hor_wind_speed(self):
        """
        Function to calculate the wind speed.
        """

        self.logger.info(
            "calculating the horizontal wind speed using DBS observations"
        )

        hor_wind_speed = np.sqrt(self.comp_v**2.0 + self.comp_u**2.0)
        hor_wind_speed.name = "hor_wind_speed"
        hor_wind_speed.attrs["long_name"] = "wind_speed"
        hor_wind_speed.attrs["units"] = "m/s"

        self.hor_wind_speed = hor_wind_speed

        return self

    def calc_hor_wind_dir(self):
        """
        Function to derive wind direction. If folows the same
        approach used by the lidar sftware.
        """

        self.logger.info("retrieving the wind direction using DBS observation")

        wind_dir = 180 + np.rad2deg(np.arctan2(-self.comp_u, -self.comp_v))

        wind_dir.name = "hor_wind_dir"
        wind_dir.attrs["long_name"] = "wind_direction"
        wind_dir.attrs["units"] = "deg"

        self.hor_wind_dir = wind_dir

        return self


class RetriveWindFFT:
    """6 beam wind retrieval

    Wind retrieval based on the FFT method for
    the 6 beam observations.

    Parameters
    ----------
    transfd_data : object
        An instance of the re-structured data, it
        should be preferentially filtered for artefacts.

    Returns
    -------
    object : object
        An object containing an dataset of the retrieved wind
        speed, direction, wind components (meridional, zonal
        and vertical) and the relative beta

    """

    def __init__(self, transfd_data: GetRestructuredData):

        self.logger = logging.getLogger(
            "lidarwind.wind_prop_retrieval.FourierTransfWindMethod"
        )
        self.logger.info("creating an instance of FourierTransfWindMethod")

        if not isinstance(transfd_data, GetRestructuredData):
            self.logger.error(
                "wrong data type: expecting a lidarwind.GetRestructuredData instance"
            )
            raise TypeError

        self.transfd_data = transfd_data
        self.ret_hor_wind_data()
        self.ret_vert_wind_data()
        self.get_beta()
        self.load_attrs()

    def ret_hor_wind_data(self):
        """
        It applies the FFT based method to retrieve
        the horizontal wind information
        """

        self.logger.info("retrieving horizontal wind from the 6 beam data")

        tmp_wind_prop = FourierTransfWindMethod(
            self.transfd_data.data_transf
        ).wind_prop()
        tmp_wind_prop = tmp_wind_prop.squeeze(dim="elv")
        tmp_wind_prop = tmp_wind_prop.drop(["elv", "freq_azm"])
        self.wind_prop = tmp_wind_prop

        return self

    def ret_vert_wind_data(self):
        """
        It copies the vertical wind from the observations
        """

        self.logger.info("selecting the vertical wind observations")

        tmp_wind_w = self.transfd_data.data_transf_90
        tmp_wind_w = tmp_wind_w.rename({"time": "time90", "range90": "range"})
        self.wind_prop["vertical_wind_speed"] = tmp_wind_w

        return self

    def get_beta(self):
        """
        It copies the raw beta from the vertical observations
        """

        self.logger.info("selcting beta from vertical observations")

        tmp_beta = self.transfd_data.relative_beta90
        tmp_beta = tmp_beta.rename({"time": "time90", "range90": "range"})
        self.wind_prop["lidar_relative_beta"] = tmp_beta

        return self

    def load_attrs(self):
        """
        It loads the attributes from all variables
        into the dataset
        """

        self.logger.info("loading data attributes")

        self.wind_prop = LoadAttributes(self.wind_prop).data

        return self



import pytest
import xarray as xr
import numpy as np

import lidarwind as lst


def test_wind_prop_retrieval_ffWindPropRet_doppler_obs():

    with pytest.raises(TypeError):
        lst.FourierTransfWindMethod(doppler_obs=np.array([0, 1]))


def test_wind_prop_retrieval_GetWindProperties5Beam_data():

    with pytest.raises(TypeError):
        lst.GetWindProperties5Beam(data=xr.DataArray(np.array([0, 1])))


def test_wind_prop_retrieval_RetriveWindFFT_tranfdData():

    with pytest.raises(TypeError):
        lst.RetriveWindFFT(transfd_data=np.array([0, 1]))



import datatree
import pandas as pd
import xarray as xr


def open_sweep(file_name):
    """Windcube's data reader

    It opens and reads the original NetCDF output
    from the Windcube lidar.

    Parameters
    ----------

    file_name : str
        path to the file that will be opened

    Returns
    -------
    ds : xarray.DataSet

        a dataset from the original NetCDF files
    """

    raw_data = datatree.open_datatree(file_name, decode_times=False)

    assert (
        "sweep_group_name" in raw_data
    ), "missing sweep group variable in input file"
    sweep_group_name = raw_data["sweep_group_name"].values[0]
    ds = raw_data[f"/{sweep_group_name}"].to_dataset()

    assert "time_reference" in ds, "missing time_reference in input sweep"
    # Guarantee that it is a valid datetime
    reference_time = pd.to_datetime(ds["time_reference"].values).isoformat()
    ds["time"].attrs["units"] = f"seconds since {reference_time}"

    ds = xr.decode_cf(ds)

    return ds



import pytest
import numpy as np
import xarray as xr

import lidarwind as lst


def test_get_dummy_six_beam_obj():

    wind = (
        np.sin(np.deg2rad(np.array([0, 72, 144, 216, 288]) + 0))
        * 1
        / np.sin(np.deg2rad(45))
    )

    elv = np.array([45, 45, 45, 45, 45, 90])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 72, 144, 216, 288, 0])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.append(wind, 0)[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    data90 = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range90"),
        coords={"time": np.arange(len(elv)), "range90": [1]},
    )

    data_status = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr90": data90,
            "gate_index90": data90,
            "radial_wind_speed90": data90,
            "radial_wind_speed_status90": data90,
            "relative_beta90": data90,
            "cnr": data,
            "gate_index": data,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "relative_beta": data,
        }
    )

    return lst.GetRestructuredData(test_ds)


@pytest.fixture
def test_get_fft_obj():

    test_ds = test_get_dummy_six_beam_obj()
    fft_obj = lst.FourierTransfWindMethod(test_ds.data_transf)

    return fft_obj


def test_fourier_transf_wind_method_pase(test_get_fft_obj):

    assert np.all(np.round(test_get_fft_obj.phase.values, 2) == 90)


def test_fourier_transf_wind_method_wind_dir(test_get_fft_obj):

    assert np.all(np.round(test_get_fft_obj.wind_dir.values, 2) == 270)


def test_fourier_transf_wind_method_rad_wind_speed(test_get_fft_obj):
    print(test_get_fft_obj.rad_wind_speed.values)
    assert np.all(
        np.round(test_get_fft_obj.rad_wind_speed.values, 2)
        == np.round(1 / np.sin(np.deg2rad(45)), 2)
    )


def test_fourier_transf_wind_method_hor_wind_speed(test_get_fft_obj):

    assert np.all(np.round(test_get_fft_obj.hor_wind_speed.values, 2) == 2)


def test_fourier_transf_wind_method_comp_u(test_get_fft_obj):

    assert np.all(np.round(test_get_fft_obj.comp_u.values, 2) == 2)


def test_fourier_transf_wind_method_comp_v(test_get_fft_obj):

    assert np.all(np.round(test_get_fft_obj.comp_v.values, 2) == 0)


def test_fourier_transf_wind_method_dim_tim(test_get_fft_obj):

    assert "time" in test_get_fft_obj.wind_prop().dims


def test_fourier_transf_wind_method_dim_range(test_get_fft_obj):

    assert "range" in test_get_fft_obj.wind_prop().dims


def test_fourier_transf_wind_method_dim_elv(test_get_fft_obj):

    assert "elv" in test_get_fft_obj.wind_prop().dims


def test_fourier_transf_wind_method_variable_wind_direction(test_get_fft_obj):

    assert "horizontal_wind_direction" in test_get_fft_obj.wind_prop()


def test_fourier_transf_wind_method_variable_wind_speed(test_get_fft_obj):

    assert "horizontal_wind_speed" in test_get_fft_obj.wind_prop()


def test_fourier_transf_wind_method_variable_zonal_wind(test_get_fft_obj):

    assert "zonal_wind" in test_get_fft_obj.wind_prop()


def test_fourier_transf_wind_method_variable_zonal_meridional(
    test_get_fft_obj,
):

    assert "meridional_wind" in test_get_fft_obj.wind_prop()



"""Utilities module

"""

import os
import shutil
import glob

import gdown
import pandas as pd
import numpy as np
import xarray as xr


class Util:

    """
    This class contains useful tools
    """

    def get_time_bins(sel_day, freq="10min"):
        """Bins estimation

        Creating time bins for a given day and time resolution

        """

        start = sel_day.strftime("%Y%m%d")
        start_time = pd.to_datetime(f"{start} 00:00:00")

        end = (sel_day + pd.to_timedelta(1, "D")).strftime("%Y%m%d")
        end_time = pd.to_datetime(f"{end} 00:00:00")

        time_bins = pd.date_range(start_time, end_time, freq=freq)

        return time_bins

    def get_sample_data(sample_path, file_type):
        """Downloading data

        It downloads the sample needed for the examples.

        """

        if file_type == "12-00":
            url = "https://drive.google.com/uc?export=download&id=1i6iX6KuZOkP_WLuPZHG5uCcvRjlWS-SU"

        if file_type == "dbs":
            url = "path"

        output = f"{sample_path}{file_type}.zip"
        gdown.download(url, output, quiet=False)

        print(f"Extracting: {output}")
        shutil.unpack_archive(output, sample_path)
        os.remove(output)

    def data_filenames():
        """Sample file list

        It searches for the sample files.
        If the files do not exist, it downloads them.

        """

        home = os.path.expanduser("~")
        sample_path = f"{home}/.lidarwindrc/sample_data/"
        file_type = "12-00"  # change to 6 beam in the future

        if os.path.isdir(sample_path):

            if os.path.isdir(f"{sample_path}{file_type}/"):
                file_list = sorted(glob.glob(f"{sample_path}{file_type}/*.nc"))

                if bool(file_list) is False:
                    Util.get_sample_data(sample_path, file_type)
                    file_list = sorted(
                        glob.glob(f"{sample_path}{file_type}/*.nc")
                    )

            else:
                Util.get_sample_data(sample_path, file_type)
                file_list = sorted(glob.glob(f"{sample_path}{file_type}/*.nc"))

        else:
            os.makedirs(sample_path)
            Util.get_sample_data(sample_path, file_type)
            file_list = sorted(glob.glob(f"{sample_path}{file_type}/*.nc"))

        return file_list


class CloudMask:

    """
    This class generates the time-height cloud mask
    and the temporal cloud mask using observations
    from lidar and ceilometer.
    """

    def __init__(self, wc_data=None, ceilo_data=None, radar_data=None):

        self.ceilo_data = ceilo_data
        self.radar_data = radar_data
        self.wc_data = wc_data

        self.call_methods()

    def call_methods(self):

        if self.ceilo_data is None or self.radar_data is None:

            self.get_time_mask(mask_type="aux")

        else:

            self.clean_ceilo()
            self.clean_radar()
            self.get_cloud_mask_2d()
            self.get_time_mask(mask_type="real")

    def clean_ceilo(self):

        positive_beta = self.ceilo_data.beta_raw.where(
            self.ceilo_data.beta_raw > 0
        )
        positive_beta = positive_beta.rolling(
            time=20, center=True, min_periods=13
        ).mean()
        positive_beta = positive_beta.rolling(
            range=10, center=True, min_periods=8
        ).mean()

        # grid interpolation: to lidar time
        self.clean_ceilo_data = positive_beta.interp(
            {"time": self.wc_data.time}
        )

    def clean_radar(self):

        positive_ze = self.radar_data.radar_equivalent_reflectivity.where(
            self.radar_data.radar_equivalent_reflectivity > 0
        )

        # grid interpolation: to lidar time, to ceilo range
        self.clean_radar_data = positive_ze.interp(
            {"time": self.wc_data.time, "range": self.clean_ceilo_data.range}
        )

    def get_cloud_mask_2d(self):

        # CEILOMETER mask
        self.clean_ceilo_data.values[
            np.isfinite(self.clean_ceilo_data.values)
        ] = 2
        self.clean_ceilo_data.values[
            ~np.isfinite(self.clean_ceilo_data.values)
        ] = 0

        # RADAR mask
        self.clean_radar_data.values[
            np.isfinite(self.clean_radar_data.values)
        ] = 1
        self.clean_radar_data.values[
            ~np.isfinite(self.clean_radar_data.values)
        ] = 0

        # final mask
        self.cloud_mask = self.clean_ceilo_data + self.clean_radar_data

    def get_time_mask(self, mask_type=None):

        if mask_type == "aux":
            print("aux mask")

            aux_cloud_mask = xr.DataArray(
                np.ones(len(self.wc_data.time)),
                dims="time",
                coords={"time": self.wc_data.time.values},
            )

            self.time_cloud_mask = aux_cloud_mask

        elif mask_type == "real":
            print("real mask")

            # 6500 is the value I defined as maximum range
            high_cloud_layer = self.cloud_mask.where(
                self.cloud_mask.range > 6500
            )

            time_cloud_mask = high_cloud_layer.sum(dim="range")

            # 1 indicates that there is a cloud above
            # the maximum range
            time_cloud_mask.values[time_cloud_mask.values > 0] = 1

            self.time_cloud_mask = time_cloud_mask

        else:
            print("mask_type not defined")



"""Module for estimating turbulence

"""
import logging

import numpy as np
import xarray as xr

from .data_operator import GetRestructuredData

module_logger = logging.getLogger("lidarwind.wind_prop_retrieval_6_beam")
module_logger.debug("loading wind_prop_retrieval_6_beam")


class SixBeamMethod:

    """6 beam method

    Implementation of the 6 beam method
    to retrieve the Reynolds stress tensor
    components based on the 6 Beam method
    developed by Sathe at all 2015.
    See: https://doi.org/10.5194/amt-8-729-2015

    Parameters
    ----------
    data : object
        an instance of the object generated by the
        lidarwind.GetRestructuredData()

    freq : int
        number of profiles used to calculate
        the variance

    freq90 : int
        number of profiles used to calculate
        the variance

    Returns
    -------
    var_comp_ds : xarray.DataSet
        a dataset of the eynolds stress tensor
        matrix elementes

    """

    def __init__(self, data, freq=10, freq90=10):

        self.logger = logging.getLogger(
            "lidarwind.wind_prop_retrieval_6_beam.SixBeamMethod"
        )
        self.logger.info("creating an instance of SixBeamMethod")

        if not isinstance(data, GetRestructuredData):
            self.logger.error(
                "wrong data type: expecting a instance of GetRestructuredData"
            )
            raise TypeError

        self.elv = data.data_transf.elv.values
        self.azm = data.data_transf.azm.values

        self.get_m_matrix()
        self.get_m_matrix_inv()
        self.radial_variances = {}
        self.calc_variances(data, freq, freq90)

        self.get_s_matrix()
        self.get_sigma()
        self.get_variance_ds()

    def get_m_matrix(self):

        """
        This method populates the coefficient matrix (M).
        Each element of M is one of the coefficients from
        equation 3 from Newman et. all 2016. The lines 0 to 4
        in M are the radial velocities coefficients from the
        non 90 deg elevation and different azimuths. Line 6
        in M has the coefficients from the radial velocity
        at 90 deg elevation.
        See: https://doi.org/10.5194/amt-9-1993-2016


        M x SIGMA = S
        """

        phis = np.append(np.ones_like(self.azm) * self.elv, np.array([90]))
        phis_rad = np.deg2rad(phis)

        thetas = np.append(self.azm, np.array([0]))
        thetas_rad = np.deg2rad(thetas)

        m_matrix = np.ones((len(phis), len(thetas))) * np.nan

        for i, theta in enumerate(thetas_rad):

            phi = phis_rad[i]

            ci1 = np.cos(phi) ** 2 * np.sin(theta) ** 2
            ci2 = np.cos(phi) ** 2 * np.cos(theta) ** 2

            ci3 = np.sin(phi) ** 2
            ci4 = np.cos(phi) ** 2 * np.cos(theta) * np.sin(theta)

            ci5 = np.cos(phi) * np.sin(phi) * np.sin(theta)
            ci6 = np.cos(phi) * np.sin(phi) * np.cos(theta)

            m_matrix_line = np.array(
                [ci1, ci2, ci3, ci4 * 2, ci5 * 2, ci6 * 2]
            )

            m_matrix[i] = m_matrix_line

        self.m_matrix = m_matrix

        return self

    def get_m_matrix_inv(self):

        """
        This method calculates the inverse matrix of M.
        """

        self.m_matrix_inv = np.linalg.inv(self.m_matrix)

        return self

    # new approach to calculate the variances ##############

    def calc_variances(self, data, freq, freq90):

        interp_data_transf = data.data_transf.interp(
            time=data.data_transf_90.time, method="nearest"
        )
        self.get_variance(interp_data_transf, freq=freq)
        self.get_variance(
            -1 * data.data_transf_90, freq=freq90, name="rVariance90"
        )  # think about the -1 coefficient

        return self

    def get_variance(self, data, freq=10, name="rVariance"):

        """
        This method calculates the variance from the
        observed radial velocities within a time window.
        The default size of this window is 10 minutes.

        Parameters
        ----------
        data : xarray.DataArray
            a dataarray of the slanted azimuthal observations

        freq : int
            number of profiles used to calculate
            the variance

        """

        variance = data.rolling(
            time=freq, center=True, min_periods=int(freq * 0.3)
        ).var()

        self.radial_variances[name] = variance

        return self

    # new approach to calculate the variances ##############

    def get_s_matrix(self):

        """
        This method fills the observation variance matrix (S).
        """

        s_matrix = np.dstack(
            (
                self.radial_variances["rVariance"].values,
                self.radial_variances["rVariance90"].values[
                    :, :, np.newaxis, np.newaxis
                ],
            )
        )

        self.s_matrix = s_matrix

    def get_sigma(self):

        """
        This method calculates the components of the
        Reynolds stress tensor (SIGMA).

        SIGMA = M^-1 x S
        """

        self.sigma_matrix = np.matmul(self.m_matrix_inv, self.s_matrix)

        return self

    def get_variance_ds(self):

        """
        This method converts the SIGMA into a xarray dataset.
        """

        var_comp_ds = xr.Dataset()
        var_comp_name = ["u", "v", "w", "uv", "uw", "vw"]

        for i, var_comp in enumerate(var_comp_name):

            tmp_data = xr.DataArray(
                self.sigma_matrix[:, :, i, 0],
                dims=("time", "range"),
                coords={
                    "time": self.radial_variances["rVariance90"].time,
                    "range": self.radial_variances["rVariance"].range,
                },
                name=f"var_{var_comp}",
            )

            var_comp_ds = xr.merge([var_comp_ds, tmp_data])

        self.var_comp_ds = var_comp_ds



import pytest
import numpy as np
import xarray as xr

import lidarwind as lst


def get_dummy_six_beam_obj():

    wind = (
        np.sin(np.deg2rad(np.array([0, 72, 144, 216, 288]) + 0))
        * 1
        / np.sin(np.deg2rad(45))
    )

    elv = np.array([45, 45, 45, 45, 45, 90])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 72, 144, 216, 288, 0])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.append(wind, 0)[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    data90 = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range90"),
        coords={"time": np.arange(len(elv)), "range90": [1]},
    )

    data_status = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr90": data90,
            "gate_index90": data90,
            "radial_wind_speed90": data90,
            "radial_wind_speed_status90": data90,
            "relative_beta90": data90,
            "cnr": data,
            "gate_index": data,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "relative_beta": data,
        }
    )

    return lst.GetRestructuredData(test_ds)


@pytest.fixture
def get_wind_profiles():

    return lst.RetriveWindFFT(get_dummy_six_beam_obj())


def test_retrieve_wind_fft_dim_elv(get_wind_profiles):
    assert "elv" not in get_wind_profiles.wind_prop.dims


def test_retrieve_wind_fft_variabe_realtive_beta(get_wind_profiles):
    assert "lidar_relative_beta" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_variable_vertical_wind(get_wind_profiles):
    assert "vertical_wind_speed" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_variable_wind_direction(get_wind_profiles):
    assert "horizontal_wind_direction" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_variable_wind_speed(get_wind_profiles):
    assert "horizontal_wind_speed" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_variable_zonal_wind(get_wind_profiles):
    assert "zonal_wind" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_variable_meridional_wind(get_wind_profiles):
    assert "meridional_wind" in get_wind_profiles.wind_prop


def test_retrieve_wind_fft_relative_beta_val(get_wind_profiles):
    assert get_wind_profiles.wind_prop["lidar_relative_beta"] == 1


def test_retrieve_wind_fft_vertical_wind_val(get_wind_profiles):
    assert get_wind_profiles.wind_prop["vertical_wind_speed"] == 1


def test_retrieve_wind_fft_global_attrs(get_wind_profiles):
    assert len(get_wind_profiles.wind_prop.attrs) > 0


def test_retrieve_wind_fft_time_len(get_wind_profiles):
    assert len(get_wind_profiles.wind_prop.time) == 5


def test_retrieve_wind_fft_time90_len(get_wind_profiles):
    assert len(get_wind_profiles.wind_prop.time90) == 1


def test_retrieve_wind_fft_range_len(get_wind_profiles):
    assert len(get_wind_profiles.wind_prop.range) == 1



import pytest
import numpy as np
import xarray as xr

from lidarwind.data_operator import wc_fixed_preprocessing
import lidarwind as lst

from .data import sample_dataset


def test_data_operator_DataOperations_data_paths():

    with pytest.raises(FileNotFoundError):
        lst.DataOperations(data_paths=None)


def test_data_operator_ReadProcessedData_file_list():

    with pytest.raises(FileNotFoundError):
        lst.ReadProcessedData(file_list=None)


def test_data_operator_GetRestructuredData_data():

    with pytest.raises(TypeError):
        lst.GetRestructuredData(data=xr.DataArray(np.array([0, 1])))


def test_data_operator_getResampled_xr_data_array_none():

    with pytest.raises(TypeError):
        lst.GetResampledData(xr_data_array=np.array([0, 1]))


def test_data_operator_DbsOperations_file_list_none():

    with pytest.raises(FileNotFoundError):
        lst.DbsOperations(file_list=None, var_list=["range"])


def test_data_operator_DbsOperations_varList_none():

    with pytest.raises(KeyError):
        lst.DbsOperations(file_list=["file_path"], var_list=None)


def test_wc_fixed_preprocessing_vertical():
    """Pre-process a vertical WC dataset"""
    ds = sample_dataset("WLS200s-218_2021-05-13_12-00-14_fixed_399_50m.nc")
    # Sanity check. Is it indeed a vertical example?
    assert (ds.elevation == 90).all()

    ds = wc_fixed_preprocessing(ds)
    assert "elevation" in ds.dims


def test_wc_fixed_preprocessing_slanted():
    """Pre-process a slanted WC dataset"""
    ds = sample_dataset("WLS200s-218_2021-05-13_12-00-08_fixed_381_50m.nc")
    # Sanity check. Is it indeed a slanted example?
    assert (ds.elevation != 90).all()

    ds = wc_fixed_preprocessing(ds)
    assert "elevation" in ds.dims


def test_wc_fixed_preprocessing_without_elevation():
    """It should raise an error with a Dataset missing elevation

    It would be best to raise a specific error in wc_fixed_preprocessing().
    """
    ds = sample_dataset("WLS200s-218_2021-05-13_12-00-08_fixed_381_50m.nc")
    ds = ds.drop("elevation")

    with pytest.raises(AssertionError):
        ds = wc_fixed_preprocessing(ds)



import pytest
import numpy as np
import xarray as xr

import lidarwind as lst


@pytest.fixture
def get_dummy_six_beam_data():

    wind = (
        np.sin(np.deg2rad(np.array([0, 72, 144, 216, 288]) + 0))
        * 1
        / np.sin(np.deg2rad(45))
    )

    elv = np.array([45, 45, 45, 45, 45, 90])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 72, 144, 216, 288, 0])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.append(wind, 0)[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    data90 = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range90"),
        coords={"time": np.arange(len(elv)), "range90": [1]},
    )

    data_status = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr90": data90,
            "gate_index90": data90,
            "radial_wind_speed90": data90,
            "radial_wind_speed_status90": data90,
            "relative_beta90": data90,
            "cnr": data,
            "gate_index": data,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "relative_beta": data,
        }
    )

    return test_ds


@pytest.fixture
def get_restruc_obj(get_dummy_six_beam_data):
    return lst.GetRestructuredData(get_dummy_six_beam_data)


def test_get_resctructured_data_range_non_90_size(get_restruc_obj):
    assert len(get_restruc_obj.range_non_90.values) == 1


def test_get_resctructured_data_range_non_90_values(get_restruc_obj):
    assert np.round(get_restruc_obj.range_non_90.values, 1) == 1


def test_get_resctructured_data_data_transf_time_dim(get_restruc_obj):
    assert "time" in get_restruc_obj.data_transf.dims


def test_get_resctructured_data_data_transf_range_dim(get_restruc_obj):
    assert "range" in get_restruc_obj.data_transf.dims


def test_get_resctructured_data_data_transf_azm_dim(get_restruc_obj):
    assert "azm" in get_restruc_obj.data_transf.dims


def test_get_resctructured_data_data_transf_elv_dim(get_restruc_obj):
    assert "elv" in get_restruc_obj.data_transf.dims


def test_get_resctructured_data_data_transf_elv_value(get_restruc_obj):
    assert np.round(get_restruc_obj.data_transf.elv.values, 1) == 45


def test_get_resctructured_data_data_transf_azm_value(get_restruc_obj):
    assert (
        np.round(get_restruc_obj.data_transf.azm.values, 1)
        == np.array([0, 72, 144, 216, 288])
    ).all()


def test_get_resctructured_data_data_transf_90_time_dim(get_restruc_obj):
    assert "time" in get_restruc_obj.data_transf_90.dims


def test_get_resctructured_data_data_transf_90_range90_dim(get_restruc_obj):
    assert "range90" in get_restruc_obj.data_transf_90.dims


def test_get_resctructured_data_data_transf_90_name(get_restruc_obj):
    assert "radial_wind_speed90" in get_restruc_obj.data_transf_90.name


def test_get_resctructured_data_data_transf_90_time_dim_size(get_restruc_obj):
    assert len(get_restruc_obj.data_transf_90.time) == 1


def test_get_resctructured_data_data_transf_90_range90_dim_size(
    get_restruc_obj,
):
    assert len(get_restruc_obj.data_transf_90.range90) == 1


def test_get_resctructured_data_input_ds_missing_cnr(get_dummy_six_beam_data):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["cnr"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds, snr=1)


def test_get_resctructured_data_input_ds_missing_cnr90(
    get_dummy_six_beam_data,
):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["cnr90"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds, snr=1)


def test_get_resctructured_data_input_ds_missing_radial_wind_speed_status90(
    get_dummy_six_beam_data,
):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["radial_wind_speed_status90"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds)


def test_get_resctructured_data_input_ds_missing_radial_wind_speed_status(
    get_dummy_six_beam_data,
):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["radial_wind_speed_status"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds)


def test_get_resctructured_data_input_ds_missing_elevation(
    get_dummy_six_beam_data,
):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["elevation"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds)


def test_get_resctructured_data_input_ds_missing_azimuth(
    get_dummy_six_beam_data,
):

    broken_ds = get_dummy_six_beam_data.copy()
    del broken_ds["azimuth"]

    with pytest.raises(AttributeError):
        lst.GetRestructuredData(broken_ds)



"""Module to manage and prepare the data for other modules

"""


import logging

import xarray as xr
import datetime as dt
import pandas as pd
import numpy as np

from .filters import Filtering

from .lidar_code import GetLidarData

module_logger = logging.getLogger("lidarwind.data_operator")
module_logger.debug("loading data_operator")


def wc_fixed_preprocessing(ds: xr.Dataset, azimuth_resolution: int = 1):

    ds["azimuth"] = ds["azimuth"].round(azimuth_resolution)
    # Avoid ambiguity on 360 degrees
    ds["azimuth"] = ds["azimuth"].where(ds.azimuth != 360, 0)

    assert "elevation" in ds
    assert ds["elevation"].dims == ("time",)
    assert ds.dims["time"] == 1

    ds["elevation"] = ds["elevation"].squeeze()
    return ds.expand_dims("elevation").set_coords("elevation")


class DataOperations:

    """Basic data manager

    It performs some basic operations. For example: rounds
    the values from the azimuth coordinate and separates
    the vertical observations from the slanted observations.

    It is helpful first use this class to process all data
    and then save the data as NetCDF files to speed up the
    derivation of wind properties later.

    Examples
    --------
    >>> merged_ds = lidarwind.DataOperations(file_list).merged_data
    >>> merged_ds.to_netcdf(output_file_path)

    Parameters
    ----------
    data_paths : list
       List of paths of the original WindCube's output.

    Returns
    -------
    object : object
        it returns an object containing an instance of the
        original files merged (.merged_data)

    """

    def __init__(self, data_paths, verbose=False):

        self.logger = logging.getLogger(
            "lidarwind.data_operator.DataOperations"
        )
        self.logger.info("creating an instance of DataOperations")

        if bool(data_paths) is False:
            self.logger.error(
                "lidarwind stopped due to an empty list of files."
            )
            raise FileNotFoundError

        self.verbose = verbose
        self.data_paths = data_paths
        self.tmp90 = xr.Dataset()
        self.tmp_non_90 = xr.Dataset()

        self.elevation_filter()
        self.rename_var_90()
        self.get_merge_data()

    def elevation_filter(self):
        """
        It groups the data from the vertical and slanted observations
        and rounds the azimuth coordinate
        """

        self.logger.info("coverting azimuth: from 360 to 0 degrees")

        for file_path in self.data_paths:

            try:
                tmp_file = GetLidarData(file_path).open_lidar_file()
                self.logger.debug(f"reading file: {file_path}")
            except:
                self.logger.warning(f"This file has a problem: {file_path}")

            try:
                elevation = tmp_file.elevation.round(1)
                tmp_file["elevation"] = elevation
                tmp_file["azimuth"] = tmp_file.azimuth.round(1)
                tmp_file["azimuth"][tmp_file.azimuth == 360] = 0
            except:
                self.logger.info(f"Problems reading elv and axm: {file_path}")

            try:
                self.tmp90 = xr.merge(
                    [self.tmp90, tmp_file.where(elevation == 90, drop=True)]
                )
            except:
                self.logger.info(
                    f"This file does not have 90 elv: {file_path}"
                )

            try:
                self.tmp_non_90 = xr.merge(
                    [
                        self.tmp_non_90,
                        tmp_file.where(elevation != 90, drop=True),
                    ]
                )
            except:
                self.logger.info(f"This file only has 90 elv: {file_path}")

        return self

    def rename_var_90(self):
        """
        It renames the vertical coordinate
        """

        self.logger.info(
            "renaming range coordinate from vertical measurements"
        )

        self.tmp90 = self.tmp90.rename_dims({"range": "range90"})

        for var in self.tmp90.variables:

            if "range90" in self.tmp90[var].dims:

                self.tmp90 = self.tmp90.rename({var: f"{var}90"})

        return self

    def get_merge_data(self):
        """
        It merges all readable data
        """

        self.logger.info("merging vertical and non-vertical measurements")

        self.merged_data = xr.merge([self.tmp90, self.tmp_non_90])

        return self


class ReadProcessedData:
    """Pre-processed data reader

    It reads all data pre-processed by data_operator.DbsOperations
    and merges them.

    Examples
    --------
    >>> merged_data = lidarwind.ReadProcessedData(file_list).merge_data()

    Parameters
    ----------
    file_list : list
        list of pre-processed NetCDF files

    Returns
    -------
    object : object
        an instance of all pre-processed data

    """

    def __init__(self, file_list):

        self.logger = logging.getLogger(
            "lidarwind.data_operator.ReadProcessedData"
        )
        self.logger.info("creating an instance of ReadProcessedData")

        if bool(file_list) is False:
            self.logger.error(
                "lidarwind stopped due to an empty list of files."
            )
            raise FileNotFoundError

        self.file_list = file_list

    def merge_data(self):
        """
        It merges all data from the file_list. It can choose between
        two different methods. One uses xr.open_mfdataset and the other
        uses xr.open_dataset.
        """

        # open_msfdataset was massing up the dimensions
        # from radial observations.

        self.logger.info("merging pre-processed data")

        try:
            tmp_merged = self.merge_data_method_1()

        except:
            print("switching from xr.open_mfdataset to xr.open_dataset")
            tmp_merged = self.merge_data_method_2()

        return tmp_merged

    def merge_data_method_1(self):
        """
        It merges data using xr.open_mfdatset
        """

        self.logger.info("mergin files using xr.open_mfdataset")

        tmp_merged = xr.open_mfdataset(self.file_list, parallel=True)

        return tmp_merged

    def merge_data_method_2(self):
        """
        It merges data using xr.open_dataset
        """

        self.logger.info("mergin files using xr.open_dataset")

        tmp_merged = xr.Dataset()

        for file_name in sorted(self.file_list):

            try:
                self.logger.info(f"opening {file_name}")
                tmp_merged = xr.merge([tmp_merged, xr.open_dataset(file_name)])

            except:
                self.logger.info(f"problems with: {file_name}")
                pass

        return tmp_merged


class GetRestructuredData:

    """Data re-structurer

    It prepares the data structure for using the wind retrieval
    modules.

    Examples
    --------
    >>> wind_prop = lidarwind.GetRestructuredData(merged_data)

    Parameters
    ----------
    data : xr.Dataset
        a xr.Dataset of pre-processed data

    snr : bool, int, optional
        if an interger is given it is used to
        as threshold to filter the data based on
        the signal to noise ratio

    status : bool, optional
        if true it filters the data using the status
        variable generated by the WindCube's software

    sProf : int, optional
        number of profiles used to calculate the anomaly for
        partially filter the second trip echoes
        (IT GOES TO FILTER MODULE)

    center : bool, optional
        (IT GOES TO FILTER MODULE)

    min_periods : int, optional
        (IT GOES TO FILTER MODULE)

    n_std : int, optional
        size of the standard deviation window used
        to partially remove the second trip echoes
        (IT GOES TO FILTER MODULE)

    check90 : bool, optional
        If True, it checks if the vertical observations
        are available. If False, the verification is ignored.

    Returns
    -------
    object : object

        an instance of the prepared for the retrieval

    """

    def __init__(
        self,
        data: xr.Dataset,
        snr=False,
        status=True,
        n_prof=500,
        center=True,
        min_periods=30,
        n_std=2,
        check90=True,
    ):

        self.logger = logging.getLogger(
            "lidarwind.data_operator.GetRestructuredData"
        )
        self.logger.info("creating an instance of GetRestructuredData")

        if not isinstance(data, xr.Dataset):
            self.logger.error("wrong data type: expecting a xr.Dataset")
            raise TypeError

        self.data = data
        self.snr = snr
        self.status = status
        self.n_prof = n_prof
        self.center = center
        self.min_periods = min_periods
        self.n_std = n_std

        self.vertical_component_check(check90)
        self.get_coord_non_90()
        self.data_transform()
        self.data_transform_90()

    def get_coord_non_90(self):

        """
        It identifies and selects the slanted data
        """

        self.logger.info("identifying and selecting the slanted observations")

        self.elv_non_90 = np.unique(
            self.data.elevation.where(self.data.elevation != 90, drop=True)
        )
        self.azm_non_90 = np.unique(
            self.data.azimuth.where(self.data.elevation != 90, drop=True)
        )
        self.azm_non_90 = np.sort(self.azm_non_90)

        self.time_non_90 = self.data.time.where(
            self.data.elevation != 90, drop=True
        )
        self.range_non_90 = self.data.range

        return self

    def data_transform(self):

        """
        It creates an xr.DataArray from all slanted observations
        """

        self.logger.info("creating a DataArray of the slanted observations")

        dop_wind_arr = np.empty(
            (
                self.time_non_90.shape[0],
                self.range_non_90.shape[0],
                len(self.azm_non_90),
                len(self.elv_non_90),
            )
        )

        for j, elv in enumerate(self.elv_non_90):

            for i, azm in enumerate(self.azm_non_90):

                tmp_rad_wind = Filtering(self.data).get_radial_obs_comp(
                    "radial_wind_speed", azm, snr=self.snr, status=self.status
                )

                dop_wind_arr[:, :, i, j] = tmp_rad_wind.sel(
                    time=self.time_non_90, method="Nearest"
                ).values

        new_range = self.data.range90.values[: len(self.data.range)]
        resampled_dop_vel = xr.DataArray(
            data=dop_wind_arr,
            dims=("time", "range", "azm", "elv"),
            coords={
                "time": self.time_non_90,
                "range": new_range,
                "azm": self.azm_non_90,
                "elv": self.elv_non_90,
            },
        )

        resampled_dop_vel.attrs = {
            "standard_name": "radial_wind_speed",
            "units": "m s-1",
            "comments": "radial wind speed vector.",
        }

        self.data_transf = resampled_dop_vel

        return self

    def data_transform_90(self):

        """
        It creates an xr.DataArray from all vertically pointing observations
        """

        self.logger.info("selcting zenith observations")

        tmp_data = Filtering(self.data).get_vertical_obs_comp(
            "radial_wind_speed90", snr=self.snr, status=self.status
        )
        tmp_data = tmp_data.isel(range90=slice(0, len(self.range_non_90)))
        self.data_transf_90 = tmp_data

        tmp_data = Filtering(self.data).get_vertical_obs_comp(
            "relative_beta90", snr=self.snr, status=self.status
        )
        tmp_data = tmp_data.isel(range90=slice(0, len(self.range_non_90)))
        self.relative_beta90 = tmp_data

        return self

    def vertical_component_check(self, check90):

        if check90:
            if not hasattr(self.data, "radial_wind_speed90"):
                raise KeyError
        else:
            print("Vertical component check was ignored.")


class GetResampledData:
    """Alternative basic data resample

    This class is used to resample the data
    into a given temporal grid.

    It mainly used internal processings of
    the package.

    Parameters
    -----------
    xr_data_array : xr.DataArray
        varaiable that will be resampled

    vert_coord : str
        name of the vertical coordinate

    time_freq : str
        size of the window e.g.: '15s'

    tolerance : int
        maximum separation from the reference

    time_coord : str
        name of the time coordinate

    Returns
    -------
    data : xr.DataArray

        time resampled variable

    """

    def __init__(
        self,
        xr_data_array: xr.DataArray,
        vert_coord="range",
        time_freq="15s",
        tolerance=10,
        time_coord="time",
    ):

        self.logger = logging.getLogger(
            "lidarwind.data_operator.GetResampledData"
        )
        self.logger.info("creating an instance of GetResampledData")

        if not isinstance(xr_data_array, xr.DataArray):
            self.logger.error("wrong data type: expecting a xr.DataArray")
            raise TypeError

        self.var_name = xr_data_array.name
        self.attrs = xr_data_array.attrs
        data = xr_data_array
        date = pd.to_datetime(data[time_coord].values[0])

        self.time_ref = self.get_time_ref(date, time_freq)
        self.vert_coord = data[vert_coord]

        time_ref_sec = np.array(self.time_ref, float) * 10 ** (-9)
        time_orig_sec = np.array(data[time_coord].values, float) * 10 ** (-9)

        delta_grid = self.calc_delt_grid(time_ref_sec, time_orig_sec)
        time_index_array = self.get_nearest_index_method_2(
            delta_grid, tolerance
        )

        self.values = self.time_resample(
            data, time_index_array, self.vert_coord
        )
        self.resampled = self.convert_to_data_array()

    def get_time_ref(self, date, time_freq="1s"):
        """
        Genetates the time reference grid used for
        resampling the data

        Parameters
        ----------
            date : Timestamp
                date for resampling (pandas Timestamp)

            dateFreq: str
                resolution of the reference grid (str, default=1s)

        Returns
        -------
        time_ref : DatetimeIndex
            time reference grid (DatetimeIndex)

        """

        self.logger.info("defining the reference time")

        start = dt.datetime(date.year, date.month, date.day, 0, 0, 0)

        end = dt.datetime(date.year, date.month, date.day, 23, 59, 59)

        time_ref = pd.date_range(start, end, freq=time_freq)

        return time_ref

    def calc_delt_grid(self, ref_grid, orig_grid):
        """
        Calculates the distance between the reference grid
        and the radar grid (time or range)

        Parameters
        ----------
        ref_grid : numpy.array
            reference grid (array[n])

        radarGrid : numpy.array
            radar grid (array[m])

        Returns
        -------
        delta_grid : numpy.array

            distance between each element from
            the reference grid to each element from the
            radar grid
        """

        self.logger.info("calculating the distance to the reference")

        tmp_grid_2_d = np.ones((len(ref_grid), len(orig_grid))) * orig_grid

        delta_grid = tmp_grid_2_d - np.reshape(ref_grid, (len(ref_grid), 1))

        return delta_grid

    def get_nearest_index_method_2(self, delta_grid, tolerance):
        """
        Identify the index of the delta_grid that fulfil
        the resampling tolerance

        Parameters
        ----------
        delta_grid : numpy.array
            output from calcRadarDeltaGrid

        tolerance : int
            tolerance distance for detecting
            the closest neighbour (time or range)

        Returns
        -------
        grid_index : np.array

            array of indexes that fulfil the resampling
            tolerance

        """

        self.logger.info("identifying index that fulfil the tolerance")

        grid_index = np.argmin(abs(delta_grid), axis=1)
        delta_grid_min = np.min(abs(delta_grid), axis=1)
        grid_index = np.array(grid_index, float)
        grid_index[delta_grid_min > tolerance] = np.nan

        return grid_index

    def time_resample(self, data, time_index_array, vert_coord):
        """
        It resamples a given radar variable using the
        time and range index calculated by get_nearest_index_method_2

        Parameters
        ----------
        var : str
            radar variable name to be resampled

        xrDataset : xarray.dataset
            xarray dataset containing the variables to
            be resampled

        timeIdexArray : np.array
            time resampling index (output from get_nearest_index_method_2)

        rangeIndexArray : np.array
            range resampling index (output from get_nearest_index_method_2)

        Returns
        -------
        resampledArr : xarray.dataArray
            time/range resampled numpy array
        """

        self.logger.info(f"time resampling of: {self.var_name}")

        resampled_time_arr = (
            np.ones((time_index_array.shape[0], self.vert_coord.shape[0]))
            * np.nan
        )

        for position, time_index in enumerate(time_index_array):

            try:
                resampled_time_arr[position] = data.values[int(time_index)]

            except:
                pass

        return resampled_time_arr

    def convert_to_data_array(self):
        """
        It creates a DataArray of the resampled data.
        """

        self.logger.info(
            f"generating the new resampled DataArray: {self.var_name}"
        )

        tmp_dt = xr.DataArray(
            self.values,
            dims=("time_ref", self.vert_coord.name),
            coords={
                "time_ref": self.time_ref,
                self.vert_coord.name: self.vert_coord.values,
            },
            name=self.var_name,
            attrs=self.attrs,
        )

        tmp_dt[self.vert_coord.name].attrs = self.vert_coord.attrs

        return tmp_dt


class DbsOperations:
    """DBS file manager

    This class extracts the variables required to
    retrieve the wind information from the DBS files.

    Parameters
    ----------
    file_list : list
        list of DBS files
    var_list : list
        list of variables to be extracted from the DBS files

    Returns
    -------
    object : object

        it returns an object containing an instance of the
        merged files (.merged_ds)

    """

    def __init__(self, file_list, var_list):

        self.logger = logging.getLogger(
            "lidarwind.data_operator.DbsOperations"
        )
        self.logger.info("creating an instance of DbsOperations")

        self.merged_ds = xr.Dataset()
        self.file_list = file_list
        self.var_list = var_list

        self.merge_data(file_list, var_list)

    def merge_data(self, file_list, var_list):
        """
        This method merges all files from a list of DBS files

        Parameters
        ----------
        file_list : list
            list of files to be merged

        var_list : list
            list of variables to be merged

        """

        self.logger.info("merging all DBS files")

        if bool(file_list) is False:
            self.logger.error(
                "lidarwind stopped due to an empty list of DBS files."
            )
            raise FileNotFoundError

        if bool(var_list) is False:
            self.logger.error(
                "lidarwind stopped due to an empty list of variable"
            )
            raise KeyError

        for file in file_list:

            try:
                file_to_merge = GetLidarData(file).open_lidar_file()
                self.logger.debug(f"reading file: {file}")
            except:
                self.logger.warning(f"This file has a problem: {file}")
                raise

            file_to_merge = self.mean_time_derivation(file_to_merge)
            # file_to_merge = self.add_mean_time(file_to_merge)

            try:
                self.merge_2_ds(file_to_merge, var_list)
            except:
                self.logger.warning(f"Merging not possible: {file}")
                # raise

    def add_mean_time(self, lidar_ds):
        """
        This method adds the mean time to each file from
        the DBS scan strategy.

        Parameters
        ----------
        lidar_ds : xarray.DataSet
            a dataset from a sequence of scans

        """

        self.logger.info("calculating the mean DBS time for each file")

        mean_time_ns = np.array(lidar_ds.time.values, np.float64).mean()
        mean_time = pd.to_datetime(
            np.ones(len(lidar_ds.time.values)) * mean_time_ns
        )
        mean_time_da = xr.DataArray(
            data=mean_time,
            dims=("time"),
            coords={"time": lidar_ds.time},
            name="scan_mean_time",
        )

        lidar_ds = lidar_ds.merge(mean_time_da)

        return lidar_ds

    def merge_2_ds(self, file_to_merge, var_list):
        """
        This method merges the variables extracted from
        the single DBS file with the storage dataset (merged_ds).

        Parameters
        ----------
        file_to_merge : xarray.DataSet
            a single file dataset

        var_list : list
            a list of variables to be merged

        """

        self.logger.info("merging single DBS file")

        for var in var_list:

            self.merged_ds = xr.merge([self.merged_ds, file_to_merge[var]])

        self.merged_ds = xr.merge(
            [self.merged_ds, file_to_merge["scan_mean_time"]]
        )

    def mean_time_derivation(self, data):

        data.azimuth.values = np.round(data.azimuth.values)
        data.azimuth.values[data.azimuth.values == 360] = 0

        azm_ref = data.azimuth.values[0]
        index_new_scan = data.ray_index.where(
            (data.elevation != 90) & (data.azimuth == azm_ref)
        )
        index_complete_scans = index_new_scan.values[
            np.isfinite(index_new_scan.values)
        ]

        if not len(data.time) in index_complete_scans:
            index_complete_scans = np.append(
                index_complete_scans, len(data.time)
            )

        groups = data.groupby_bins(
            "ray_index", index_complete_scans, right=False
        )

        new_data = xr.Dataset()
        for grp in groups.groups:

            new_data = xr.merge([new_data, self.add_mean_time(groups[grp])])

        return new_data



import pytest

import lidarwind as lst
from .data import data_filenames, get_sample_data


@pytest.fixture
def test_DataOperations(data_filenames):

    tmp_object = lst.DataOperations(data_filenames)

    return tmp_object.merged_data


def test_GetRestructuredData(test_DataOperations):

    restruct_data = lst.GetRestructuredData(test_DataOperations)



"""Module for configuring the data global attribute

"""

import json


class Configurations:
    """Global attributes definition

    This class defines all global attributes
    that will be writen on the dataset

    Parameters
    ----------
    lidarwind : object
        an instance of the lidarwind package

    """

    def __init__(self, lidarwind=None):

        self.load_version(lidarwind)
        self.load_reference()
        self.load_institution()
        self.load_instrument()
        self.load_site()
        self.load_contact()
        self.load_email()
        self.load_comments()

    def load_version(self, lidarwind):
        """
        It identifies the lidarwind version
        and writes it to the configuration file

        Parameters
        ----------
        lidarwind : object
            a instance of the lidarwind package

        """

        if lidarwind is None:
            self.lidarwind_version = "temporary config file"
        else:
            self.lidarwind_version = lidarwind.__version__

        return self

    def load_institution(self, institution="institution name"):
        """
        It defines the institution affiliation name

        Parameters
        ----------
        institution : str
            institution name

        """

        self.institution = institution

        return self

    def load_instrument(self, instrument="instrument name"):
        """
        It defines the instrument name

        Parameters
        ----------
        instrument : str
            name of the instrument used during the experiment

        """

        self.instrument = instrument

        return self

    def load_site(self, site="site name"):
        """
        It defines the name of the experimental site

        Parameters
        ----------
        site : str
            name of the experimental site

        """

        self.site = site

        return self

    def load_contact(self, contact="contact person"):
        """
        It defines the author's name

        Parameters
        ----------
        contact : str
            name of the contact person

        """

        self.contact = contact

        return self

    def load_email(self, email="contact email"):
        """
        It defines the contacting email

        Parameters
        ----------
        email : str
            contact email

        """
        self.email = email

        return self

    def load_reference(self, reference="Generated by lidarwind version: {0}"):
        """
        It loads the lidarwind's version used for
        processing the data

        Parameters
        ----------
        reference : str
            lidarwind version used to process the data

        """

        self.references = reference.format(self.lidarwind_version)

        return self

    def load_comments(self, comments="General comments"):
        """
        It defines additional comments

        Parameters
        ----------
        comments : str
            additional comments

        """
        self.comments = comments

        return self

    def generate_conf(self):
        """
        It writes and saves all defined global attributes.

        """

        config_dic = {}

        config_dic["references"] = self.references
        config_dic["institution"] = self.institution
        config_dic["instrument_name"] = self.instrument
        config_dic["site_name"] = self.site
        config_dic["comments"] = self.comments
        config_dic["contact_person"] = self.contact
        config_dic["email"] = self.email

        config_js = json.dumps(config_dic)
        config_file = open("config.json", "w")
        config_file.write(config_js)
        config_file.close()

    def load_conf_file(self, file_path="config.json"):
        """
        It loads the pre-defined global attributes
        from the config.json, if it exists.

        Parameters
        ----------
        file_path : str
            the path to the configuration file (config.json)

        """

        try:
            config_dic = json.load(open(file_path))

        except FileNotFoundError:

            print("You do not have a config file yet")
            print("a temporary config file was generated")
            print("See the documentation for generating it")
            self.generate_conf()
            config_dic = json.load(open(file_path))

        self.load_reference(config_dic["references"])
        self.load_institution(config_dic["institution"])
        self.load_instrument(config_dic["instrument_name"])
        self.load_comments(config_dic["comments"])
        self.load_site(config_dic["site_name"])
        self.load_contact(config_dic["contact_person"])
        self.load_email(config_dic["email"])

        return self



import os
import shutil
import glob
from typing import Optional

import pytest
import gdown

from lidarwind.io import open_sweep

GDRIVE_ID = "1i6iX6KuZOkP_WLuPZHG5uCcvRjlWS-SU"


def lidarwindrc(subdir: Optional[str] = None):
    """Standard path for Lidar Suit configurations

    This might be moved into lidarwind.utils if used somewhere else
    """
    path = os.getenv("LIDARWIND_DIR", os.path.join("~", ".lidarwindrc"))
    path = os.path.expanduser(path)

    if subdir is not None:
        path = os.path.join(path, subdir)

    return path


def get_sample_data(sample_path, file_type):

    if file_type == "12-00":
        url = f"https://drive.google.com/uc?export=download&id={GDRIVE_ID}"

    if file_type == "dbs":
        url = "path"

    output = f"{sample_path}{file_type}.zip"
    gdown.download(url, output, quiet=False)

    print(f"Extracting: {output}")
    shutil.unpack_archive(output, sample_path)
    os.remove(output)


def download_samples():
    file_type = "12-00"  # change to 6 beam in the future
    sample_path = lidarwindrc("sample_data")

    if not os.path.isdir(sample_path):
        os.makedirs(sample_path)
        get_sample_data(sample_path, file_type)


@pytest.fixture
def data_filenames():

    file_type = "12-00"  # change to 6 beam in the future
    sample_path = os.path.join(lidarwindrc("sample_data"), file_type)

    if os.path.isdir(sample_path):

        if os.path.isdir(sample_path):
            file_list = sorted(glob.glob(f"{sample_path}/*.nc"))

            if bool(file_list) is False:
                get_sample_data(sample_path, file_type)
                file_list = sorted(glob.glob(f"{sample_path}/*.nc"))

        else:
            get_sample_data(sample_path, file_type)
            file_list = sorted(glob.glob(f"{sample_path}/*.nc"))

    else:
        download_samples()
        file_list = sorted(glob.glob(f"{sample_path}/*.nc"))

    return file_list


def sample_dataset(key: str):
    """Single xr.Dataset for testing

    For a given identifier, for now it is the filename, download if needed,
    and return already as an xarray Dataset ready to be used.

    !!ATENTION!!! the data_filenames as an argument is a requirement while
    the download of sample data is not refactored. This function only access
    a file that was expected to have been previously downlaoded.
    """
    path = os.path.join(lidarwindrc("sample_data"), "12-00", key)

    ds = open_sweep(path)
    return ds



"""Module to open the lidar original output

"""


import warnings


from .io import open_sweep


class GetLidarData:
    """Windcube's data reader


    ATENTION, please move to io.open_sweep(). This GetLidarData will
    be eventually removed.

    It opens and reads the original NetCDF output from the Windcube lidar

    Parameters
    ----------

    file_name : str
        name of the file that will be open

    """

    def __init__(self, file_name):

        self.file_name = file_name

    def open_lidar_file(self):

        """
        Function to read the lidar NetCDF files

        Returns
        -------
        ds: xarray.DataSet
            a dataset from the original NetCDF files

        Note
        ----
        Alias to io.open_sweep() so upper functions don't break until we
        finished the transition to new I/O.
        """
        warnings.warn(
            "GetLidarData will be removed eventually. "
            "Please use io module instead",
            DeprecationWarning,
            stacklevel=2,
        )

        return open_sweep(self.file_name)



"""Top-level package for lidarwind package"""

__author__ = "José Dias Neto"
__email__ = "jdiasn@gmail.com"
__affiliation__ = "Delft University of Technology"


from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    try:
        from .version import version as __version__
    except ImportError:
        raise ImportError(
            "Failed to find (autogenerated) version.py. "
            "This might be because you are installing from GitHub's tarballs, "
            "use the PyPI ones."
        )


from .lidarwind_config import *
from .lidar_code import *
from .wind_prop_retrieval import *
from .wind_prop_retrieval_6_beam import *
from .visualization import *
from .utilities import *
from .data_operator import *
from .data_attributes import *
from .filters import *



"""Module for wrinting the data attributes

"""

from .lidarwind_config import Configurations


class LoadAttributes:

    """Level 1 Attribute generator

    It defines and writes the attributes from the Level 1

    Parameters
    ----------
    data : xarray.Dataset
        A xarray dataset containing the Level 1 data
        generated by the lidarwind

    Returns
    -------
    data : xarray.Dataset

        A dataset with corrected attributes

    """

    def __init__(self, data):

        self.data = data
        self.write_global_attrs()
        self.variables_attrs()
        self.write_coords_attrs()
        self.write_variables_attrs()

    def write_global_attrs(self):

        """Global attribute writer

        It loads and writes the global attributes from the dataset.
        The global attributes are defiened in the configuration file.
        See lidarwind_config.generate_conf documentation for information
        about generating the configuration file.

        """

        config_info = Configurations(lidarwind=None).load_conf_file()

        tmp_att = {
            "Conventions": "Cf/Radial 2.0",
            "title": "Wind properties",
            "references": config_info.references,
            "institution": config_info.institution,
            "instrument_name": config_info.instrument,
            "comments": config_info.comments,
            "site_name": config_info.site,
            "contact_person": config_info.contact,
            "email": config_info.email,
        }

        self.data.attrs = tmp_att

        return self

    def variables_attrs(self):

        """Variable attributes definitions

        All variables and coordinates attributes are defined here.

        """

        attrs_dic = {}

        attrs_dic["range"] = {
            "standard_name": "range",
            "units": "m",
            "comments": "Distance between the instrument "
            "and the center of each range gate",
        }

        attrs_dic["time"] = {
            "standard_name": "time",
            "reference": "seconds since 1970-01-01 00:00:00",
            "comments": "time of the horizotal observations",
        }

        attrs_dic["time90"] = {
            "standard_name": "time90",
            "reference": "seconds since 1970-01-01 00:00:00",
            "comments": "time of the vertical observations",
        }

        attrs_dic["horizontal_wind_speed"] = {
            "standard_name": "wind_speed",
            "units": "m/s",
            "comments": "horizontal wind speed retrived using the FFT method",
        }

        attrs_dic["horizontal_wind_direction"] = {
            "standard_name": "wind_direction",
            "units": "degrees",
            "comments": "horizontal wind direction retrived "
            "using the FFT method with respect to true north",
            "info": "0=wind coming from the north, "
            "90=east, 180=south, 270=west",
        }

        attrs_dic["zonal_wind"] = {
            "standard_name": "zonal_wind",
            "units": "m/s",
            "comments": "zonal wind retrived using the FFT method",
        }

        attrs_dic["meridional_wind"] = {
            "standard_name": "meridional_wind",
            "units": "m/s",
            "comments": "meridional wind retrived using the FFT method",
        }

        attrs_dic["vertical_wind_speed"] = {
            "standard_name": "vertical_wind_speed",
            "units": "m/s",
            "comments": "observed vertical wind speed "
            "(negative towards the ground)",
        }

        attrs_dic["lidar_relative_beta"] = {
            "standard_name": "volume_attenuated_backwards"
            "_scattering_function_in_air",
            "units": "m-1 sr-1",
            "comments": "Attenuated relative backscatter "
            "coefficient from the vertical beam",
        }

        self.attrs_dic = attrs_dic

        return self

    def write_coords_attrs(self):

        """Coordinate attribute writer

        It loops over the dataset coordinate and writes their attributes

        """

        for key in self.data.coords:

            try:
                self.data[key].attrs = self.attrs_dic[key]

            except KeyError:
                print(f"coord not found: {key}")

        return self

    def write_variables_attrs(self):

        """Variable attribute writer

        It loops over the dataset variables and writes their attributes

        """

        for key in self.data.keys():

            try:
                self.data[key].attrs = self.attrs_dic[key]

            except KeyError:
                print(f"variable not found: {key}")

        return self



import pytest
import numpy as np
import xarray as xr

import lidarwind as lst


def test_six_beam_method_input():

    with pytest.raises(TypeError):
        lst.SixBeamMethod(data=xr.DataArray(np.array([0, 1])))


def test_get_dummy_six_beam_obj():

    elv = np.array([75, 75, 90, 75, 75, 75])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 72, 0, 144, 216, 288])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range"),
        coords={"time": np.arange(len(elv)), "range": [1]},
    )

    data90 = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "range90"),
        coords={"time": np.arange(len(elv)), "range90": [1]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr90": data90,
            "gate_index90": data90,
            "radial_wind_speed90": data90,
            "radial_wind_speed_status90": data90,
            "relative_beta90": data90,
            "cnr": data,
            "gate_index": data,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data,
            "relative_beta": data,
        }
    )

    return lst.GetRestructuredData(test_ds)


@pytest.fixture
def test_get_six_beam_obj():

    six_beam_obj = lst.SixBeamMethod(
        test_get_dummy_six_beam_obj(), freq=6, freq90=6
    )

    return six_beam_obj


def test_six_beam_method_m_matrix(test_get_six_beam_obj):

    assert np.all(np.isfinite(test_get_six_beam_obj.m_matrix))


def test_six_beam_method_m_matrix_inv(test_get_six_beam_obj):

    assert np.all(np.isfinite(test_get_six_beam_obj.m_matrix_inv))


def test_six_beam_method_variance_dic(test_get_six_beam_obj):

    assert len(test_get_six_beam_obj.radial_variances.keys()) == 2


def test_six_beam_method_radial_variances90(test_get_six_beam_obj):

    assert np.all(
        test_get_six_beam_obj.radial_variances["rVariance90"].values == 0
    )


def test_six_beam_method_radial_variances(test_get_six_beam_obj):

    assert np.all(
        test_get_six_beam_obj.radial_variances["rVariance"].values == 0
    )


def test_six_beam_method_sigma_matrix(test_get_six_beam_obj):

    assert np.all(test_get_six_beam_obj.sigma_matrix == 0)


def test_six_beam_method_variance_dim_time(test_get_six_beam_obj):

    assert len(test_get_six_beam_obj.var_comp_ds.time.values) == 1


def test_six_beam_method_variance_dim_range(test_get_six_beam_obj):

    assert len(test_get_six_beam_obj.var_comp_ds.range.values) == 1



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages


setup(
    include_package_data=True,
    packages=find_packages(include=["lidarwind", "lidarwind.*"]),
    zip_safe=False,
)
